__copyright__ = "Copyright (C) 2013-2017  Martin Blais"
__license__ = "GNU GPLv2"

import warnings
warnings.warn("beancount.ops.prices is deprecated; has moved to beancount.core.prices")

from beancount.core.prices import *  # pylint: disable=wildcard-import,unused-wildcard-import
